/**
 * Display an XML document in a JTree control.  Derived from "DOMTree.java" by
 * Andy Clark of IBM.  The original DOMTree is copyright 1999-2000 by the
 * Apache Software Foundation.  The text of the Apache Public License should
 * have been included with this distribution in the file &quot;APL.txt&quot;.
 *
 * The portions of this code not covered by the Apache Public License are
 * covered by the FiveSight Public License.  The FiveSight Public License is
 * compatible with the Apache Public License.
 *
 * ============================================================================
 * The FiveSight Public Software License
 *
 * Copyright (c) 2001 FiveSight Technologies, Inc..  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *       FiveSight Technologies, Inc. (http://www.fivesight.com/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "FiveSight" and "FiveSight Technologies" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact info@fivesight.com.
 *
 * 5. Products derived from this software may not be called "FiveSight",
 *    nor may "FiveSight" appear in their name, without prior written
 *    permission of FiveSight Technologies.
 *
 * ============================================================================
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FIVESIGHT TECHNOLOGIES, INC. BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ============================================================================
 */

package com.fs.xpathtester;

import java.io.Serializable;
import java.awt.Component;
import javax.swing.JTree;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import org.w3c.dom.*;
import java.util.Hashtable;

public class DOMTree extends JTree {

  /**
   * Whether or not to display attributes as separate nodes in the default configuration.
   */
  private static final boolean ATTRIBUTE_DISPLAY_DEFAULT = true;
  /**
   * Whether or not to display comments as separate nodes in the default configuration.
   */
  private static final boolean COMMENT_DISPLAY_DEFAULT = true;
  /**
   * Whether or not to display processing instruction nodes in the default configuration.
   */
  private static final boolean PI_DISPLAY_DEFAULT = true;


  /**
   * Constructs a tree with the specified document and settings for display
   * of attributes and comments.
   *
   * @param document the node to use in the construction of the DOMTree.
   * @param showAtts whether or not to display attributes as separate nodes in the tree
   * @param showComms whether or not to display comments as separate nodes in the tree
   */
  public DOMTree(Node document,boolean showAtts, boolean showComms, boolean showPI) {
    super(new Model(showAtts,showComms,showPI));
    this.addTreeExpansionListener( new DOMExpansionListener());
    this.putClientProperty("JTree.lineStyle", "Horizontal");
    // set properties
    setDocument(document);
    setCellRenderer(new DefaultTreeCellRenderer(){
      public Component getTreeCellRendererComponent( JTree tree, Object value,
        boolean sel, boolean expanded, boolean leaf,  int row,
        boolean hasFocus)
      {
        super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf,
          row, hasFocus);
        setIcon(null);
        return this;
      }
    });
  }

  /**
   * Create a new, empty DOMTree.
   */
  public DOMTree() { this(null); }

  /**
   * Create a new, empty DOMTree with attribute dislpay and comment display set
   * by the supplied parameters.
   *
   * @param showAtts whether or not to display attributes as separate nodes in the tree
   * @param showComms whether or not to display comments as nodes in the tree
   * @param showPI whether or not to display PI as nodes in the tree
   */
  public DOMTree(boolean showAtts, boolean showComms, boolean showPI) {
    this(null,showAtts,showComms,showPI);
  }

  /**
   * Create a new DOMTree using the supplied document.
   *
   * @param document the node to use in the construction of the DOMTree.
   */
  public DOMTree(Node document) {
    this(document,ATTRIBUTE_DISPLAY_DEFAULT,COMMENT_DISPLAY_DEFAULT,PI_DISPLAY_DEFAULT);
  }

  /**
   * Change the document being displayed.
   *
   * @param document the node containing the document (or fragment) to display.
   */
  public void setDocument(Node document) {
    ((Model)getModel()).setDocument(document);
    setRootVisible(document != null);
    expandRow(0);
  }

  /**
   * Returns the document currently being displayed.
   *
   * @return the currently displayed document as an org.w3c.dom.Node
   */
  public Node getDocument() { return ((Model)getModel()).getDocument(); }

  /**
   * Get the org.w3c.dom.Node corresponding to a MutableTreeNode.
   *
   * @param treeNode the MutableTreeNode to look up.
   * @return the Node corresponding to treeNode (null if none).
   */
  public Node getNode(Object treeNode) {
    return ((Model)getModel()).getNode(treeNode);
  }

  /**
   * Get the MutableTreeNode corresponding to a specific org.w3c.dom.Node.
   *
   * @param node the org.w3c.dom.Node to loo up
   * @return the MutableTreeNode corresponding to the DOM Node (null if none).
   */
  public MutableTreeNode getTreeNode(Node node){
    return ((Model)getModel()).getTreeNode(node);
  }

  /**
   * This TreeExpansionListener receives notification about pieces of the DOMTree
   * to expose based on XPath matches.
   */
  class DOMExpansionListener implements TreeExpansionListener {
    DefaultMutableTreeNode treeNode, sibling;
    Model model;
    Node node;

    public DOMExpansionListener() {
      model = (Model)DOMTree.this.getModel();
    }

    public void treeCollapsed(TreeExpansionEvent tee) {
      TreePath path = tee.getPath();
      treeNode = (DefaultMutableTreeNode)path.getLastPathComponent();
      sibling = treeNode.getNextSibling();
      Node node = model.getNode(treeNode);
      if(node != null && node.getNodeType() == Node.ELEMENT_NODE
        && sibling != null && model.getNode(sibling) == null)
      {
        model.removeNodeFromParent(sibling);
      }
    }

    public void treeExpanded(TreeExpansionEvent tee) {
      TreePath path = tee.getPath();
      treeNode = (DefaultMutableTreeNode)path.getLastPathComponent();
      Node node = model.getNode(treeNode);
      if(node != null && node.getNodeType() == Node.ELEMENT_NODE) {
        model.insertNodeInto(
          new DefaultMutableTreeNode("</ " + node.getNodeName() + " >"),
            (MutableTreeNode)treeNode.getParent(),
            model.getIndexOfChild(treeNode.getParent(), treeNode) + 1);
      }
    }
  }

  /**
   * TreeModel for the DOMTree.
   */
  public static class Model extends DefaultTreeModel implements Serializable {
    private Node document;
    private Hashtable nodeMap = new Hashtable();
    private Hashtable treeMap = new Hashtable();

    private boolean _showAttributesAsNodes = false;
    private boolean _showCommentsAsNodes = false;
    private boolean _showPIAsNodes = false;

    public Model() { this(null,false,false,false); }

    public Model(Node document) { this(document,false,false,false); }

    public Model(boolean showAtts, boolean showComms, boolean showPI) {
      this(null,showAtts,showComms,showPI);
    }

    public Model(Node document, boolean showAtts, boolean showComments,
      boolean showPI)
    {
      super(new DefaultMutableTreeNode());
      setDocument(document);
      _showAttributesAsNodes = showAtts;
      _showCommentsAsNodes = showComments;
      _showPIAsNodes = showPI;
    }

    /**
     * Sets the document.
     */
    public synchronized void setDocument(Node document) {
      this.document = document;
      ((DefaultMutableTreeNode)getRoot()).removeAllChildren();
      nodeMap.clear();
      buildTree();
      fireTreeStructureChanged(this, new Object[] { getRoot() }, new int[0], new Object[0]);
    }

    /**
     * Returns the document.
     */
    public Node getDocument() {
      return document;
    }

    /**
     * Get the org.w3c.dom.Node for a MutableTreeNode.
     */
    public Node getNode(Object treeNode) {
      return (Node)nodeMap.get(treeNode);
    }

    public MutableTreeNode getTreeNode(Node node){
      return (MutableTreeNode)treeMap.get(node);
    }

    /**
     * Build the tree.  Handle the supplied Node in an appropriate fashion.
     */
    private void buildTree() {
      if (document == null) { return; }
      MutableTreeNode root = (MutableTreeNode)getRoot();
      if (document.getNodeType() == Node.DOCUMENT_NODE) {
        NodeList nl = document.getChildNodes();
        for (int i=0; i < nl.getLength(); ++i) insertNode(nl.item(i),root);
      } else if (document.getNodeType() == Node.ELEMENT_NODE)
        insertElementNode(document,root);
      else if (document.getNodeType() == Node.TEXT_NODE)
        insertTextNode(document,root);
      else if (document.getNodeType() == Node.CDATA_SECTION_NODE)
        insertCDataSectionNode(document,root);
      else if (document.getNodeType() == Node.COMMENT_NODE) {
        _showCommentsAsNodes = true;
        insertCommentNode(document,root);
      } else if (document.getNodeType() == Node.ATTRIBUTE_NODE) {
        _showAttributesAsNodes = true;
        insertAttributeNode(document,root);
      }
    }

    /**
     * Switch method for inserting a node based on type.
     */
    private MutableTreeNode insertNode(Node n, MutableTreeNode where) {
      switch (n.getNodeType()) {
        case Node.TEXT_NODE: return insertTextNode(n,where);
        case Node.ELEMENT_NODE: return insertElementNode(n,where);
        case Node.ATTRIBUTE_NODE: return insertAttributeNode(n,where);
        case Node.COMMENT_NODE: return insertCommentNode(n,where);
        case Node.PROCESSING_INSTRUCTION_NODE: return insertPINode(n,where);
        case Node.CDATA_SECTION_NODE: return insertCDataSectionNode(n,where);
        default: return null;
      }
    }

    /**
     * Insert a node and returns a reference to the new node.
     */
    private MutableTreeNode insertNode(String what, MutableTreeNode where) {
      MutableTreeNode node = new DefaultMutableTreeNode(what);
      insertNodeInto(node, where, where.getChildCount());
      return node;
    }

    /**
     * Insert an element node.
     */
    private MutableTreeNode insertElementNode(Node what, MutableTreeNode where) {
      StringBuffer name = new StringBuffer();
      NamedNodeMap attrs = what.getAttributes();
      int attrCount = (attrs != null) ? attrs.getLength() : 0;
      name.append("< ");
      name.append(what.getNodeName());
      if (_showAttributesAsNodes) {
        if (attrCount>0) {
          name.append(" (expand for attributes)");
        }
      } else {
        for (int i=0;i < attrCount; ++i) {
          name.append(" ");
          name.append(attrs.item(i).getNodeName());
          name.append("=\"");
          name.append(attrs.item(i).getNodeValue());
          name.append("\"");
        }
      }
      if(what.hasChildNodes()) name.append(" >");
      else name.append(" />");
      MutableTreeNode element = insertNode(name.toString(), where);
      nodeMap.put(element, what);
      treeMap.put(what, element);
      for (int i=0;i<attrCount;++i) insertAttributeNode(attrs.item(i),element);
      NodeList children = what.getChildNodes();
      int len = (children != null) ? children.getLength() : 0;
      for (int i = 0; i < len; i++) insertNode(children.item(i),element);
      return element;
    }

    /**
     * Insert a text node.
     */
    private MutableTreeNode insertTextNode(Node what, MutableTreeNode where) {
      String value = (_showAttributesAsNodes ? "[#text: "
        + what.getNodeValue().trim() + " ]"
        : what.getNodeValue().trim());

      if (value.length() > 0) {
        MutableTreeNode treeNode = insertNode(value, where);
        nodeMap.put(treeNode, what);
        treeMap.put(what, treeNode);
        return treeNode;
      }
      return null;
    }

    /**
     * Insert a comment node.
     */
    private MutableTreeNode insertCommentNode(Node what, MutableTreeNode where) {
      if (_showCommentsAsNodes) {
        String value = "<!-- " + what.getNodeValue().trim() + " -->";
        MutableTreeNode treeNode = insertNode(value, where);
        nodeMap.put(treeNode, what);
        treeMap.put(what, treeNode);
        return treeNode;
      } else {
        return null;
      }
    }

    /**
     * Insert a PI node.
     */
    private MutableTreeNode insertPINode(Node what, MutableTreeNode where) {
      if (_showPIAsNodes) {
        String value="<?" + what.getNodeName() + " " + what.getNodeValue() + " ?>";
        MutableTreeNode treeNode = insertNode(value,where);
        nodeMap.put(treeNode,what);
        treeMap.put(what,treeNode);
        return treeNode;
      } else {
        return null;
      }
    }

    /**
     * Insert a node for an attribute.
     */
    private MutableTreeNode insertAttributeNode(Node what, MutableTreeNode where) {
      if (_showAttributesAsNodes) {
        StringBuffer attrbuff = new StringBuffer();
        String name = what.getNodeName();
        String value = what.getNodeValue();
        if (value != null) {
          attrbuff.append("@");
          attrbuff.append(name);
          attrbuff.append("=\"");
          attrbuff.append(value);
          attrbuff.append("\"");
          MutableTreeNode treeNode = insertNode(attrbuff.toString(),where);
          nodeMap.put(treeNode,what);
          treeMap.put(what,treeNode);
          return treeNode;
        }
      } else {
        // if we're not supposed to display attributes separately, hard-wire the
        // attribute into the parent element for highlighting purposes.
        nodeMap.put(where,what);
        treeMap.put(what,where);
        return where;
      }
      return null;
    }

    /**
     * Insert a CDATA Section Node.
     */
    private MutableTreeNode insertCDataSectionNode(Node what, MutableTreeNode where) {
      StringBuffer CSectionBfr = new StringBuffer();
      CSectionBfr.append("<![CDATA[");
      CSectionBfr.append(what.getNodeValue());
      CSectionBfr.append("]]>");
      MutableTreeNode treeNode = insertNode(CSectionBfr.toString(), where);
      nodeMap.put(treeNode, what);
      treeMap.put(what, treeNode);
      return treeNode;
    }
  }
}
